import os
from typing import Any, Set, Tuple

import logging

from regipy.exceptions import RegistryKeyNotFoundException

from regipy.registry import RegistryHive, NKRecord
from regipy.utils import convert_wintime, calculate_sha1

logger = logging.getLogger(__name__)


def get_subkeys_and_timestamps(registry_hive):
    subkeys_and_timestamps = set()
    for subkey in registry_hive.recurse_subkeys():
        subkey_path = subkey.path
        ts = subkey.timestamp
        subkeys_and_timestamps.add((subkey_path, ts))
    return subkeys_and_timestamps


def get_values_from_tuples(value_tuples, value_name_list):
    for value_name, value_data in value_tuples:
        if value_name in value_name_list:
            yield value_name, value_data


def get_timestamp_for_subkeys(registry_hive, subkey_list):
    for subkey_path in subkey_list:
        try:
            subkey = registry_hive.get_key(subkey_path)
        except RegistryKeyNotFoundException as ex:
            logger.exception(f'Could not obtain timestamp for subkey {subkey_path}')
            continue
        yield subkey_path, convert_wintime(subkey.header.last_modified, as_json=True)


def _get_name_value_tuples(subkey: NKRecord) -> Set[Tuple[str, Any]]:
    """
    Iterate over value in a subkey and return a set of tuples containing value names and values
    :param subkey: NKRecord to iterate over
    :return: A set of tuples containing value names and values
    """
    values_tuple: Set[Tuple[str, Any]] = set()
    for value in subkey.iter_values(as_json=True):
        if not value.value:
            continue

        if isinstance(value.value, list):
            values_tuple.update({(value.name, x) for x in value.value})
        else:
            values_tuple.add((value.name, value.value))
    return values_tuple


def compare_hives(first_hive_path, second_hive_path, verbose=False):
    # The list will contain tuples, in the following format: (Difference type, first value, second value, description)
    found_differences = []

    # Compare hash, verify they are indeed different
    first_hive_sha1 = calculate_sha1(first_hive_path)
    second_hive_sha1 = calculate_sha1(second_hive_path)

    if first_hive_sha1 == second_hive_sha1:
        logger.info('Hives have the same hash!')
        return found_differences

    # Compare header parameters
    first_registry_hive = RegistryHive(first_hive_path)
    second_registry_hive = RegistryHive(second_hive_path)
    if first_registry_hive.header.hive_bins_data_size != second_registry_hive.header.hive_bins_data_size:
        found_differences.append(('different_hive_bin_data_size', first_registry_hive.header.hive_bins_data_size,
                                  second_registry_hive.header.hive_bins_data_size, ''))

    # Enumerate subkeys for each hive and start comparing
    logger.info('Enumerating subkeys in {}'.format(os.path.basename(first_hive_path)))
    first_hive_subkeys = get_subkeys_and_timestamps(first_registry_hive)

    logger.info('Enumerating subkeys in {}'.format(os.path.basename(second_hive_path)))
    second_hive_subkeys = get_subkeys_and_timestamps(second_registry_hive)

    # Get a set of keys present in one hive and not the other and vice versa
    first_hive_subkey_names = {x[0] for x in first_hive_subkeys if x[0] is not None}
    second_hive_subkey_names = {x[0] for x in second_hive_subkeys if x[0] is not None}

    found_differences.extend(('new_subkey', ts, None, subkey_path) for subkey_path, ts in
                             get_timestamp_for_subkeys(first_registry_hive,
                                                       first_hive_subkey_names - second_hive_subkey_names))

    found_differences.extend(('new_subkey', None, ts, subkey_path) for subkey_path, ts in
                             get_timestamp_for_subkeys(second_registry_hive,
                                                       second_hive_subkey_names - first_hive_subkey_names))

    # Remove duplicate keys from each of the sets
    first_hive_diff_subkeys = first_hive_subkeys - second_hive_subkeys
    second_hive_diff_subkeys = second_hive_subkeys - first_hive_subkeys

    # Find subkeys that exist in both hives, but were modified. Look for new values and subkeys
    for path_1, ts_1 in first_hive_diff_subkeys:
        for path_2, ts_2 in second_hive_diff_subkeys:
            if path_1 and path_1 == path_2 and ts_1 != ts_2:
                first_subkey_nk_record = first_registry_hive.get_key(path_1)
                second_subkey_nk_record = second_registry_hive.get_key(path_2)

                # Compare values between the subkeys
                first_subkey_values = set()
                second_subkey_values = set()

                if first_subkey_nk_record.values_count:
                    first_subkey_values = _get_name_value_tuples(first_subkey_nk_record)

                if second_subkey_nk_record.values_count:
                    second_subkey_values = _get_name_value_tuples(second_subkey_nk_record)

                # If one hive or the other contain values, and they are different, compare values
                if (first_subkey_values or second_subkey_values) and (first_subkey_values != second_subkey_values):
                    first_hive_value_names = set(x[0] for x in first_subkey_values)
                    second_hive_value_names = set(x[0] for x in second_subkey_values)

                    values_in_first_but_not_in_second = first_hive_value_names - second_hive_value_names
                    values_in_second_but_not_in_first = second_hive_value_names - first_hive_value_names

                    # If there are value names that are present in the first subkey but not the second
                    # Iterate over all values in the first subkey
                    # If the value name is one of those that is not on the second subkey, add it to the set
                    if values_in_first_but_not_in_second:
                        found_differences.extend(('new_value', f'{n}: {d} @ {ts_1}', None, path_1) for n, d in
                                                 get_values_from_tuples(first_subkey_values,
                                                                        values_in_first_but_not_in_second))

                    if values_in_second_but_not_in_first:
                        found_differences.extend(('new_value', None, f'{n}: {d} @ {ts_2}', path_1) for n, d in
                                                 get_values_from_tuples(second_subkey_values,
                                                                        values_in_second_but_not_in_first))

                # We do not compare subkeys for each subkey, because we would have detected those.
    return found_differences



import logging
from regipy.exceptions import RegistryKeyNotFoundException
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime
import re

logger = logging.getLogger(__name__)

NTUSER_SHELLBAG = '\\Software\\Microsoft\\Windows\\Shell\\BagMRU'
USRCLASS_SHELLBAG = '\\Local Settings\\Software\\Microsoft\\Windows\\Shell\\BagMRU'
CODEPAGE = 'cp1252'


class ShellBagNtuserPlugin(Plugin):
    NAME = 'shellbag_plugin'
    DESCRIPTION = 'Parse Shellbag items'
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    @staticmethod
    def _parse_mru(mru_val):
        mru_order_string = ''
        if isinstance(mru_val, bytes):
            mru_val = mru_val[:-4]
            for i in range(0, len(mru_val), 4):
                mru_order_string += f'{str(mru_val[i])}-'

            return mru_order_string[:-1]
        else:
            return mru_order_string

    @staticmethod
    def _get_shell_item_type(shell_item):
        try:
            import pyfwsi
        except ModuleNotFoundError as ex:
            logger.exception(f"Plugin `shellbag_plugin` has missing modules, install regipy using"
                             f" `pip install regipy[full]` in order to install plugin dependencies. "
                             f"This might take some time... ")
            raise ex

        if isinstance(shell_item, pyfwsi.volume):
            item_type = "Volume"

        elif isinstance(shell_item, pyfwsi.file_entry):
            item_type = "Directory"

        elif isinstance(shell_item, pyfwsi.network_location):
            item_type = "Network Location"

        else:
            item_type = 'unknown'

        return item_type

    @staticmethod
    def _parse_shell_item_path_segment(shell_item):
        """Parses a shell item path segment.
        Args:
          shell_item (pyfwsi.item): shell item.
        Returns:
          str: shell item path segment.
        """

        try:
            import pyfwsi
        except ModuleNotFoundError as ex:
            logger.exception(f"Plugin `shellbag_plugin` has missing modules, install regipy using"
                             f" `pip install regipy[full]` in order to install plugin dependencies. "
                             f"This might take some time... ")
            raise ex

        path_segment = None

        if isinstance(shell_item, pyfwsi.volume):
            if shell_item.name:
                path_segment = shell_item.name
            elif shell_item.identifier:
                path_segment = '{{{0:s}}}'.format(shell_item.identifier)

        elif isinstance(shell_item, pyfwsi.file_entry):
            long_name = ''
            for extension_block in shell_item.extension_blocks:
                if isinstance(extension_block, pyfwsi.file_entry_extension):
                    long_name = extension_block.long_name

            if long_name:
                path_segment = long_name
            elif shell_item.name:
                path_segment = shell_item.name

        elif isinstance(shell_item, pyfwsi.network_location):
            if shell_item.location:
                path_segment = shell_item.location

        if path_segment is None:
            path_segment = '<UNKNOWN: 0x{0:02x}>'.format(shell_item.class_type)

        return path_segment

    def iter_sk(self, key, reg_path, base_path='', path=''):
        try:
            import pyfwsi
        except ModuleNotFoundError as ex:
            logger.exception(f"Plugin `shellbag_plugin` has missing modules, install regipy using"
                             f" `pip install regipy[full]` in order to install plugin dependencies. "
                             f"This might take some time... ")
            raise ex

        last_write = convert_wintime(key.header.last_modified, as_json=True)

        mru_val = key.get_value('MRUListEx')
        mru_order = self._parse_mru(mru_val)
        base_path = path

        if key.get_value('NodeSlot'):
            node_slot = str(key.get_value('NodeSlot'))
        else:
            node_slot = ''

        for v in key.iter_values(trim_values=False):
            if re.match("\d+", v.name):
                slot = v.name
                byte_stream = v.value
                shell_items = pyfwsi.item_list()
                shell_items.copy_from_byte_stream(byte_stream, ascii_codepage=CODEPAGE)
                for item in shell_items.items:
                    shell_type = self._get_shell_item_type(item)
                    value = self._parse_shell_item_path_segment(item)
                    if shell_type != 'unknown':
                        if not path:
                            path = value
                            base_path += f'\\{value}' if base_path else value
                        else:
                            path += f'\\{value}'

                    creation_time = None
                    access_time = None
                    modification_time = None

                    if len(item.extension_blocks) > 0:
                        for extension_block in item.extension_blocks:
                            if isinstance(extension_block, pyfwsi.file_entry_extension):
                                try:
                                    creation_time = extension_block.get_creation_time()
                                    if self.as_json:
                                        creation_time = creation_time.isoformat()
                                except OSError:
                                    logger.exception(f'Malformed creation time for {path}')
                                try:
                                    access_time = extension_block.get_access_time()
                                    if self.as_json:
                                        access_time = access_time.isoformat()
                                except OSError:
                                    logger.exception(f'Malformed access time for {path}')

                    try:
                        if hasattr(item, 'modification_time'):
                            modification_time = item.get_modification_time()
                            if self.as_json:
                                modification_time = modification_time.isoformat()
                    except OSError:
                        logger.exception(f'Malformed modification time for {path}')

                    value_name = v.name
                    mru_order_location = mru_order.split('-').index(value_name)
                    entry = {'value': value,
                             'slot': slot,
                             'reg_path': reg_path,
                             'value_name': value_name,
                             'node_slot': node_slot,
                             'shell_type': shell_type,
                             'path': path,
                             'creation_time': creation_time,
                             'access_time': access_time,
                             'modification_time': modification_time,
                             'last_write': last_write,
                             'mru_order': mru_order,
                             'mru_order_location': mru_order_location,
                             }

                    self.entries.append(entry)
                    sk_reg_path = f'{reg_path}\\{value_name}'
                    sk = self.registry_hive.get_key(sk_reg_path)
                    self.iter_sk(sk, sk_reg_path, base_path, path)
                    path = base_path

    def run(self):

        try:
            import pyfwsi
        except ModuleNotFoundError as ex:
            logger.exception(f"Plugin `shellbag_plugin` has missing modules, install regipy using"
                             f" `pip install regipy[full]` in order to install plugin dependencies. "
                             f"This might take some time... ")
            raise ex

        try:
            shellbag_ntuser_subkey = self.registry_hive.get_key(NTUSER_SHELLBAG)
            self.iter_sk(shellbag_ntuser_subkey, NTUSER_SHELLBAG)
        except RegistryKeyNotFoundException as ex:
            logger.error(f'Could not find {self.NAME} plugin data at: {NTUSER_SHELLBAG}: {ex}')


class RegipyException(Exception):
    """
    This is the parent exception for all regipy exceptions
    """
    pass

class RegipyGeneralException(RegipyException):
    """
    General exception
    """
    pass


class RegistryValueNotFoundException(RegipyException):
    pass


class NoRegistrySubkeysException(RegipyException):
    pass


class NoRegistryValuesException(RegipyException):
    pass


class RegistryKeyNotFoundException(RegipyException):
    pass


class UnidentifiedHiveException(RegipyException):
    pass


class RegistryRecoveryException(RegipyException):
    pass


class RegistryParsingException(RegipyException):
    """
    Raised when there is a parsing error, most probably a corrupted hive
    """
    pass

class NtSidDecodingException(RegipyException):
    """
    Raised when the binary Windows NT SID representation can not be decoded
    """



import binascii
import datetime as dt

from typing import List, Optional, Union

import logging

import attr

from construct import Bytes, CString, EnumIntegerString, GreedyRange, Int32sl, Int32ul, Int64ul, \
    StreamError, ConstError
from io import BytesIO


from regipy.exceptions import NoRegistrySubkeysException, RegistryKeyNotFoundException, NoRegistryValuesException, \
    RegistryValueNotFoundException, RegipyGeneralException, UnidentifiedHiveException, RegistryParsingException
from regipy.hive_types import SUPPORTED_HIVE_TYPES
from regipy.security_utils import convert_sid, get_acls
from regipy.structs import REGF_HEADER, HBIN_HEADER, CM_KEY_NODE, LF_LH_SK_ELEMENT, VALUE_KEY, INDEX_ROOT, \
    REGF_HEADER_SIZE, INDEX_ROOT_SIGNATURE, LEAF_INDEX_SIGNATURE, FAST_LEAF_SIGNATURE, HASH_LEAF_SIGNATURE, \
    BIG_DATA_BLOCK, INDEX_LEAF, DEFAULT_VALUE, SECURITY_KEY_v1_1, SECURITY_DESCRIPTOR, SID, ACL, ACE, VALUE_TYPE_ENUM
from regipy.utils import boomerang_stream, convert_wintime, identify_hive_type, MAX_LEN, try_decode_binary

logger = logging.getLogger(__name__)


@attr.s
class Cell:
    """
    Represents a Registry cell header
    """
    offset: int = attr.ib()
    cell_type: str = attr.ib()
    size: int = attr.ib()


@attr.s
class VKRecord:
    """
    The VK Record contains a value
    """
    value_type: EnumIntegerString = attr.ib()
    value_type_str: str = attr.ib()
    value: bytes = attr.ib()
    size: int = attr.ib(default=0)
    is_corrupted: bool = attr.ib(default=False)


@attr.s
class LIRecord:
    data: bytes = attr.ib()


@attr.s
class Value:
    name: str = attr.ib()
    value: Union[str, int, bytes] = attr.ib()
    value_type: str = attr.ib()
    is_corrupted: bool = attr.ib(default=False)


@attr.s
class Subkey:
    subkey_name: str = attr.ib()
    path: str = attr.ib()
    timestamp: dt.datetime = attr.ib()
    values_count: int = attr.ib()
    values: List[Value] = attr.ib(factory=list)

    # This field will be used if a partial hive was given, if not it would be None.
    actual_path: Optional[str] = attr.ib(default=None)


class RIRecord:
    data = None
    header = None

    def __init__(self, stream):
        self.header = INDEX_ROOT.parse_stream(stream)


class RegistryHive:
    CONTROL_SETS = [r'\ControlSet001', r'\ControlSet002']

    def __init__(self, hive_path, hive_type=None, partial_hive_path=None):
        """
        Represents a registry hive
        :param hive_path: Path to the registry hive
        :param hive_type: The hive type can be specified if this is a partial hive,
                          or for some other reason regipy cannot identify the hive type
        :param partial_hive_path: The path from which the partial hive actually starts, for example:
                                  hive_type=ntuser partial_hive_path="/Software" would mean
                                  this is actually a HKCU hive, starting from HKCU/Software
        """

        self.partial_hive_path = None
        self.hive_type = None

        with open(hive_path, 'rb') as f:
            self._stream = BytesIO(f.read())

        with boomerang_stream(self._stream) as s:
            self.header = REGF_HEADER.parse_stream(s)

            # Get the first cell in root HBin, which is the root NKRecord:
            root_hbin = self.get_hbin_at_offset()
            root_hbin_cell = next(root_hbin.iter_cells(s))
            self.root = NKRecord(root_hbin_cell, s)
        self.name = self.header.file_name

        if hive_type:
            if hive_type.lower() in SUPPORTED_HIVE_TYPES:
                self.hive_type = hive_type
            else:
                raise UnidentifiedHiveException(f'{hive_type} is not a supported hive type: '
                                                f'only the following are supported: {SUPPORTED_HIVE_TYPES}')
        else:
            try:
                self.hive_type = identify_hive_type(self.name)
            except UnidentifiedHiveException:
                logger.info(f'Hive type for {hive_path} was not identified: {self.name}')

        if partial_hive_path:
            self.partial_hive_path = partial_hive_path

    def recurse_subkeys(self, nk_record=None, path_root=None, as_json=False, is_init=True, fetch_values=True):
        """
        Recurse over a subkey, and yield all of its subkeys and values
        :param nk_record: an instance of NKRecord from which to start iterating, if None, will start from Root
        :param path_root: If we are iterating an incomplete hive, for example a hive tree starting
                          from ControlSet001 and not SYSTEM, there is no way to know that.
                          This string will be added as prefix to all paths.
        :param as_json: Whether to normalize the data as JSON or not
        :param fetch_values: If False, subkey values will not be returned, but the iteration will be faster
        """
        # If None, will start iterating from Root NK entry
        if not nk_record:
            nk_record = self.root

        # Iterate over subkeys
        if nk_record.header.subkey_count:
            for subkey in nk_record.iter_subkeys():
                if path_root:
                    subkey_path = r'{}\{}'.format(path_root, subkey.name) if path_root else r'\{}'.format(subkey.name)
                else:
                    subkey_path = f'\\{subkey.name}'

                # Leaf Index records do not contain subkeys
                if isinstance(subkey, LIRecord):
                    continue

                if subkey.subkey_count:
                    yield from self.recurse_subkeys(nk_record=subkey,
                                                    path_root=subkey_path,
                                                    as_json=as_json,
                                                    is_init=False,
                                                    fetch_values=fetch_values)

                values = []
                if fetch_values:
                    if subkey.values_count:
                        try:
                            if as_json:
                                values = [attr.asdict(x) for x in subkey.iter_values(as_json=as_json)]
                            else:
                                values = list(subkey.iter_values(as_json=as_json))
                        except RegistryParsingException as ex:
                            logger.exception(f'Failed to parse hive value at path: {path_root}')

                ts = convert_wintime(subkey.header.last_modified)
                yield Subkey(subkey_name=subkey.name, path=subkey_path,
                             timestamp=ts.isoformat() if as_json else ts, values=values,
                             values_count=subkey.values_count,
                             actual_path=f'{self.partial_hive_path}{subkey_path}' if self.partial_hive_path else None)

        if is_init:
            # Get the values of the subkey
            values = []
            if nk_record.values_count:
                try:
                    if as_json:
                        values = [attr.asdict(x) for x in nk_record.iter_values(as_json=as_json)]
                    else:
                        values = list(nk_record.iter_values(as_json=as_json))
                except RegistryParsingException as ex:
                    logger.exception(f'Failed to parse hive value at path: {path_root}')
                    values = []

            ts = convert_wintime(nk_record.header.last_modified)
            subkey_path = path_root or '\\'
            yield Subkey(subkey_name=nk_record.name, path=subkey_path,
                         timestamp=ts.isoformat() if as_json else ts, values=values, values_count=len(values),
                         actual_path=f'{self.partial_hive_path}\\{subkey_path}' if self.partial_hive_path else None)

    def get_hbin_at_offset(self, offset=0):
        """
        This offset is from start of data (meaning that it starts from 4096)
        If not offset is given, will return the root Hbin
        :return:
        """
        self._stream.seek(REGF_HEADER_SIZE + offset)
        return HBin(self._stream)

    def get_key(self, key_path):
        if self.partial_hive_path:
            if key_path.startswith(self.partial_hive_path):
                key_path = key_path.partition(self.partial_hive_path)[-1]
            else:
                raise RegistryKeyNotFoundException(f'Did not find subkey at {key_path}, because this is a partial hive')

        logger.debug('Getting key: {}'.format(key_path))

        # If the key path is \ we are just refering to root
        if key_path == '\\':
            return self.root

        # If the path contain slashes, this is a full path. Split it
        if '\\' in key_path:
            key_path_parts = key_path.split('\\')[1:]
        else:
            key_path_parts = [key_path]

        previous_key_name = []

        subkey = self.root.get_subkey(key_path_parts.pop(0), raise_on_missing=False)

        if not subkey:
            raise RegistryKeyNotFoundException('Did not find subkey at {}'.format(key_path))

        if not key_path_parts:
            return subkey

        for path_part in key_path_parts:
            new_path = '\\'.join(previous_key_name)
            previous_key_name.append(subkey.name)
            subkey = subkey.get_subkey(path_part, raise_on_missing=False)

            if not subkey:
                raise RegistryKeyNotFoundException('Did not find {} at {}'.format(path_part, new_path))
        return subkey

    def get_control_sets(self, registry_path):
        """
        Get the optional control sets for a registry hive
        :param registry_path:
        :return: A list of paths, including the control sets
        """

        found_control_sets = []
        for cs in self.CONTROL_SETS:
            try:
                found_control_sets.append(self.get_key(cs))
            except RegistryKeyNotFoundException:
                continue
        result = [r'\{}\{}'.format(subkey.name, registry_path) for subkey in found_control_sets]
        logger.info('Found control sets: {}'.format(result))
        return result


class HBin:
    def __init__(self, stream):
        """
        :param stream: a stream at the start of the hbin block
        """
        self.header = HBIN_HEADER.parse_stream(stream)
        self.hbin_data_offset = stream.tell()

    def iter_cells(self, stream):
        stream.seek(self.hbin_data_offset)
        offset = stream.tell()
        while offset < self.hbin_data_offset + self.header.size - HBIN_HEADER.sizeof():
            hbin_cell_size = Int32sl.parse_stream(stream)

            # If the cell size is positive, it means it is unallocated. We are not interested in those on a regular run
            if hbin_cell_size >= 0:
                continue

            bytes_to_read = (hbin_cell_size * -1) - 4

            cell_type = Bytes(2).parse_stream(stream)

            # Yield the cell
            yield Cell(cell_type=cell_type.decode(), offset=stream.tell(), size=bytes_to_read)

            # Go to the next cell
            offset += stream.tell() + bytes_to_read


class NKRecord:
    """
    The NKRecord represents a Name Key entry
    """

    def __init__(self, cell, stream):
        stream.seek(cell.offset)
        self.header = CM_KEY_NODE.parse_stream(stream)
        self._stream = stream

        # Sometimes the key names are ASCII and sometimes UTF-16 little endian
        if self.header.flags.KEY_COMP_NAME:
            # Compressed (ASCII) key name
            self.name = self.header.key_name_string.decode('ascii', errors='replace')
        else:
            # Unicode (UTF-16) key name
            self.name = self.header.key_name_string.decode('utf-16-le', errors='replace')
            logger.debug(f'Unicode key name identified: "{self.name}"')

        self.subkey_count = self.header.subkey_count
        self.values_count = self.header.values_count
        self.volatile_subkeys_count = self.header.volatile_subkey_count

    def get_subkey(self, key_name, raise_on_missing=True):
        if not self.subkey_count and raise_on_missing:
            raise NoRegistrySubkeysException('No subkeys for {}'.format(self.header.key_name_string))

        for subkey in self.iter_subkeys():
            # This should not happen
            if not isinstance(subkey, NKRecord):
                raise RegipyGeneralException(f'Unknown record type: {subkey}')

            if subkey.name.upper() == key_name.upper():
                return subkey

        if raise_on_missing:
            raise NoRegistrySubkeysException('No subkey {} for {}'.format(key_name, self.header.key_name_string))

    def iter_subkeys(self):

        if not self.header.subkey_count:
            return None

        # Go to the offset where the subkey list starts (+4 is because of the cell header)
        target_offset = REGF_HEADER_SIZE + 4 + self.header.subkeys_list_offset
        self._stream.seek(target_offset)

        # Read the signature
        try:
            signature = Bytes(2).parse_stream(self._stream)
        except StreamError as ex:
            raise RegistryParsingException(f'Bad subkey at offset {target_offset}: {ex}')

        # LF,  LH and RI contain subkeys
        if signature in [HASH_LEAF_SIGNATURE, FAST_LEAF_SIGNATURE, LEAF_INDEX_SIGNATURE]:
            yield from self._parse_subkeys(self._stream, signature=signature)
        # RI contains pointers to arrays of subkeys
        elif signature == INDEX_ROOT_SIGNATURE:
            ri_record = RIRecord(self._stream)
            if ri_record.header.element_count > 0:
                for element in ri_record.header.elements:
                    # We skip 6 because of the signature as well as the cell header
                    element_target_offset = REGF_HEADER_SIZE + 4 + element.subkey_list_offset
                    self._stream.seek(element_target_offset)
                    yield from self._parse_subkeys(self._stream)

    @staticmethod
    def _parse_subkeys(stream, signature=None):
        """
        Parse an LI , LF or LH Record
        :param stream: A stream at the header of the LH or LF entry, skipping the signature
        :return:
        """
        if not signature:
            signature = stream.read(2)

        if signature in [HASH_LEAF_SIGNATURE, FAST_LEAF_SIGNATURE]:
            subkeys = LF_LH_SK_ELEMENT.parse_stream(stream)
        elif signature == LEAF_INDEX_SIGNATURE:
            subkeys = INDEX_LEAF.parse_stream(stream)
        else:
            raise RegistryParsingException(f'Expected a known signature, got: {signature} at offset {stream.tell()}')

        for subkey in subkeys.elements:
            stream.seek(REGF_HEADER_SIZE + subkey.key_node_offset)

            # This cell should always be allocated, therefor we expect a negative size
            cell_size = Int32sl.parse_stream(stream) * -1

            # We read to this offset and skip 2 bytes, because that is the cell size we just read
            nk_cell = Cell(cell_type='nk', offset=stream.tell() + 2, size=cell_size)
            nk_record = NKRecord(cell=nk_cell, stream=stream)
            yield nk_record

    @staticmethod
    def read_value(vk, stream):
        """
        Read a registry value
        :param vk: A parse VK record
        :param stream: The registry stream
        :return: A VKRecord
        """
        stream.seek(REGF_HEADER_SIZE + 4 + vk.data_offset)
        data_type = vk.data_type
        data = stream.read(vk.data_size)
        return VKRecord(value_type=data_type, value_type_str=str(data_type), value=data, size=vk.data_size)

    @staticmethod
    def _parse_indirect_block(stream, value):
        # This is an indirect datablock (Bigger than 16344, therefor we handle it differently)
        # The value inside the vk entry actually contains a pointer to the buffers containing the data
        big_data_block_header = BIG_DATA_BLOCK.parse(value.value)

        # Go to the start of the segment offset list
        stream.seek(REGF_HEADER_SIZE + big_data_block_header.offset_to_list_of_segments)
        buffer = BytesIO()

        # Read them sequentially until we got all the size of the VK
        value_size = value.size
        while value_size > 0:
            data_segment_offset = Int32ul.parse_stream(stream)
            with boomerang_stream(stream) as tmpstream:
                tmpstream.seek(REGF_HEADER_SIZE + 4 + data_segment_offset)
                tmpbuffer = tmpstream.read(min(0x3fd8, value_size))
                value_size -= len(tmpbuffer)
                buffer.write(tmpbuffer)
        buffer.seek(0)
        return buffer.read()

    def iter_values(self, as_json=False, max_len=MAX_LEN, trim_values=True):
        """
        Get the values of a subkey. Will raise if no values exist
        :param as_json: Whether to normalize the data as JSON or not
        :param max_len: Max length of value to return
        :param trim_values: whether to trim values to MAX_LEN
        :return: List of values for the subkey
        """
        if not self.values_count:
            return

        # Get the offset of the values key. We skip 4 because of Cell Header
        target_offset = REGF_HEADER_SIZE + 4 + self.header.values_list_offset
        self._stream.seek(target_offset)

        for _ in range(self.values_count):
            is_corrupted = False
            try:
                vk_offset = Int32ul.parse_stream(self._stream)
            except StreamError:
                logger.info(f'Skipping bad registry VK at {self._stream.tell()}')
                raise RegistryParsingException(f'Bad registry VK at {self._stream.tell()}')

            with boomerang_stream(self._stream) as substream:
                actual_vk_offset = REGF_HEADER_SIZE + 4 + vk_offset
                substream.seek(actual_vk_offset)
                try:
                    vk = VALUE_KEY.parse_stream(substream)
                except (ConstError, StreamError):
                    logger.error(f'Could not parse VK at {substream.tell()}, registry hive is probably corrupted.')
                    return

                value = self.read_value(vk, substream)

                if vk.name_size == 0:
                    value_name = '(default)'
                elif vk.flags.VALUE_COMP_NAME:
                    # Compressed (ASCII) value name
                    value_name = vk.name.decode('ascii', errors='replace')
                else:
                    # Unicode (UTF-16) value name
                    value_name = vk.name.decode('utf-16-le', errors='replace')
                    logger.debug(f'Unicode value name identified: "{value_name}"')

                # If the value is bigger than this value, it means this is a DEVPROP structure
                # https://doxygen.reactos.org/d0/dba/devpropdef_8h_source.html
                # https://sourceforge.net/p/mingw-w64/mingw-w64/ci/668a1d3e85042c409e0c292e621b3dc0aa26177c/tree/
                # mingw-w64-headers/include/devpropdef.h?diff=dd86a3b7594dadeef9d6a37c4b6be3ca42ef7e94
                # We currently do not support these, We are going to make the best effort to dump as string.
                # This int casting will always work because the data_type is construct's EnumIntegerString
                if int(vk.data_type) > 0xffff0000:
                    data_type = VALUE_TYPE_ENUM.parse(Int32ul.build(int(vk.data_type) & 0xffff))
                    logger.info(f"Value at {hex(actual_vk_offset)} contains DEVPROP structure of type {data_type}")

                # Skip this unknown data type, research pending :)
                # TODO: Add actual parsing
                elif int(vk.data_type) == 0x200000:
                    logger.info(f"Skipped unknown data type value at {actual_vk_offset}")
                    continue
                else:
                    data_type = str(vk.data_type)

                if data_type in ['REG_SZ', 'REG_EXPAND', 'REG_EXPAND_SZ']:
                    if vk.data_size >= 0x80000000:
                        # data is contained in the data_offset field
                        value.size -= 0x80000000
                        actual_value = vk.data_offset
                    elif vk.data_size > 0x3fd8 and value.value[:2] == b'db':
                        data = self._parse_indirect_block(substream, value)
                        actual_value = try_decode_binary(data, as_json=as_json, trim_values=trim_values)
                    else:
                        actual_value = try_decode_binary(value.value, as_json=as_json, trim_values=trim_values)
                elif data_type in ['REG_BINARY', 'REG_NONE']:
                    if vk.data_size >= 0x80000000:
                        # data is contained in the data_offset field
                        actual_value = vk.data_offset
                    elif vk.data_size > 0x3fd8 and value.value[:2] == b'db':
                        try:
                            actual_value = self._parse_indirect_block(substream, value)

                            actual_value = try_decode_binary(actual_value, as_json=True, trim_values=trim_values) if as_json else actual_value
                        except ConstError:
                            logger.error(f'Bad value at {actual_vk_offset}')
                            continue
                    else:
                        # Return the actual data
                        actual_value = binascii.b2a_hex(value.value).decode()[:max_len] if trim_values else value.value
                elif data_type == 'REG_SZ':
                    actual_value = try_decode_binary(value.value, as_json=as_json, trim_values=trim_values)
                elif data_type == 'REG_DWORD':
                    # If the data size is bigger than 0x80000000, data is actually stored in the VK data offset.
                    actual_value = vk.data_offset if vk.data_size >= 0x80000000 else Int32ul.parse(value.value)
                elif data_type == 'REG_QWORD':
                    actual_value = vk.data_offset if vk.data_size >= 0x80000000 else Int64ul.parse(value.value)
                elif data_type == 'REG_MULTI_SZ':
                    parsed_value = GreedyRange(CString('utf-16-le')).parse(value.value)
                    # Because the ListContainer object returned by Construct cannot be turned into a list,
                    # we do this trick
                    actual_value = [x for x in parsed_value if x]
                # We currently dumps this as hex string or raw
                # TODO: Add actual parsing
                elif data_type in ['REG_RESOURCE_REQUIREMENTS_LIST', 'REG_RESOURCE_LIST']:
                    actual_value = binascii.b2a_hex(value.value).decode()[:max_len] if trim_values else value.value
                elif data_type == 'REG_FILETIME':
                    actual_value = convert_wintime(Int64ul.parse(value.value), as_json=as_json)
                else:
                    actual_value = try_decode_binary(value.value, as_json=as_json, trim_values=trim_values)
                yield Value(name=value_name, value_type=data_type, value=actual_value,
                            is_corrupted=is_corrupted)

    def get_value(self, value_name=DEFAULT_VALUE, as_json=False, raise_on_missing=False, case_sensitive=True):
        """
        Get a value by name. Will raise if raise_on_missing is set,
        if no value name is given, will return the content of the default value
        :param value_name: The value name to look for
        :param as_json: Whether to normalize the data as JSON or not
        :param raise_on_missing: Will raise exception if value is missing, else will return None
        :return:
        """
        value_name = value_name if case_sensitive else value_name.lower()
        for value in self.iter_values(as_json=as_json, trim_values=False):
            v = value.name if case_sensitive else value.name.lower()
            if v == value_name:
                return value.value

        if raise_on_missing:
            raise RegistryValueNotFoundException('Did not find the value {} on subkey {}'.format(value_name, self.name))
        return None

    def get_values(self, as_json=False):
        return [x for x in self.iter_values(as_json=as_json)]

    def get_security_key_info(self):
        self._stream.seek(REGF_HEADER_SIZE + self.header.security_key_offset)
        # TODO: If parsing fails, parse with SECURITY_KEY_v1_2
        security_key = SECURITY_KEY_v1_1.parse_stream(self._stream)
        security_descriptor = SECURITY_DESCRIPTOR.parse(security_key.security_descriptor)

        with boomerang_stream(self._stream) as s:
            security_base_offset = REGF_HEADER_SIZE + self.header.security_key_offset + 24

            s.seek(security_base_offset + security_descriptor.owner)
            owner_sid = convert_sid(SID.parse_stream(s))

            s.seek(security_base_offset + security_descriptor.group)
            group_sid = convert_sid(SID.parse_stream(s))

            sacl_aces = None
            if security_descriptor.offset_sacl > 0:
                s.seek(security_base_offset + security_descriptor.offset_sacl)
                sacl_aces = get_acls(s)

            dacl_aces = None
            if security_descriptor.offset_dacl > 0:
                s.seek(security_base_offset + security_descriptor.offset_dacl)
                dacl_aces = get_acls(s)
            return {
                'owner': owner_sid,
                'group': group_sid,
                'dacl': dacl_aces,
                'sacl': sacl_aces
            }

    def get_class_name(self) -> str:
        """
        Gets the key class name as would be returned via
        the `lpClass` argument of the `RegQueryInfoKey()` function.
        """

        # Get the offset of the class name string. We skip 4 because of Cell Header
        read_offset = REGF_HEADER_SIZE + 4 + self.header.class_name_offset

        self._stream.seek(read_offset)
        class_name = self._stream.read(self.header.class_name_size)

        return class_name.decode('utf-16-le', errors='replace')

    def __dict__(self):
        return {
            'name': self.name,
            'subkey_count': self.subkey_count,
            'value_count': self.values_count,
            'values': {x['name']: x['value'] for x in self.iter_values()} if self.values_count else None,
            'subkeys': {x['name'] for x in self.iter_subkeys()} if self.subkey_count else None,
            'timestamp': convert_wintime(self.header.last_modified, as_json=True),
            'volatile_subkeys': self.volatile_subkeys_count
        }



NTUSER_HIVE_TYPE = 'ntuser'
SYSTEM_HIVE_TYPE = 'system'
AMCACHE_HIVE_TYPE = 'amcache'
SOFTWARE_HIVE_TYPE = 'software'
SAM_HIVE_TYPE = 'sam'
SECURITY_HIVE_TYPE = 'security'
CLASSES_ROOT_HIVE_TYPE = 'classes_root'
BCD_HIVE_TYPE = 'bcd'
USRCLASS_HIVE_TYPE = 'usrclass'


SUPPORTED_HIVE_TYPES = [
    NTUSER_HIVE_TYPE,
    SYSTEM_HIVE_TYPE,
    AMCACHE_HIVE_TYPE,
    SOFTWARE_HIVE_TYPE,
    SAM_HIVE_TYPE,
    SECURITY_HIVE_TYPE,
    BCD_HIVE_TYPE,
    USRCLASS_HIVE_TYPE
]


HVLE_TRANSACTION_LOG_MAGIC = b'HvLE'
DIRT_TRANSACTION_LOG_MAGIC = b'DIRT'



import logging

from construct import Int64ul

from regipy.exceptions import RegistryKeyNotFoundException
from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

BAM_PATH = r'Services\bam\UserSettings'


class BAMPlugin(Plugin):
    NAME = 'background_activity_moderator'
    DESCRIPTION = 'Get the computer name'
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        logger.info('Started Computer Name Plugin...')

        try:
            for subkey_path in self.registry_hive.get_control_sets(BAM_PATH):
                subkey = self.registry_hive.get_key(subkey_path)
                for sid_subkey in subkey.iter_subkeys():

                    sid = sid_subkey.name
                    logger.info(f'Parsing BAM for {sid}')
                    sequence_number = None
                    version = None
                    entries = []

                    for value in sid_subkey.get_values():
                        if value.name == 'SequenceNumber':
                            sequence_number = value.value
                        elif value.name == 'Version':
                            version = value.value
                        else:
                            entries.append({
                                'executable': value.name,
                                'timestamp': convert_wintime(Int64ul.parse(value.value), as_json=self.as_json)
                            })

                    self.entries.extend([
                        {
                            'sequence_number': sequence_number,
                            'version': version,
                            'sid': sid,
                            **x
                        } for x in entries]
                    )
        except RegistryKeyNotFoundException as ex:
            logger.error(ex)



import logging

from regipy import RegistryKeyNotFoundException, convert_wintime
from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin

logger = logging.getLogger(__name__)

UAC_KEY_PATH = r'\Microsoft\Windows\CurrentVersion\Policies\System'


class UACStatusPlugin(Plugin):
    NAME = 'uac_plugin'
    DESCRIPTION = 'Get the status of User Access Control'
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def run(self):
        """
        This plugin checks the following parameters:
        EnableLUA - Windows notifies the user when programs try to make changes to the computer
        EnableVirtualization - enables the redirection of legacy application File and Registry writes
                               that would normally fail as standard user to a user-writable data location.
        FilterAdministratorToken - If enabled approval is required when performing administrative tasks.
        ConsentPromptBehaviorAdmin - This option allows the Consent Admin to perform an
                                     operation that requires elevation without consent or credentials.
        ConsentPromptBehaviorUser - If enabled,  a standard user that needs to perform an operation that requires
                                    elevation of privilege will be prompted for an administrative user name and password
        """
        try:
            subkey = self.registry_hive.get_key(UAC_KEY_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(f'Could not find {self.NAME} subkey at {UAC_KEY_PATH}: {ex}')
            return None

        self.entries = {
            'last_write': convert_wintime(subkey.header.last_modified, as_json=self.as_json),
            'enable_limited_user_accounts': subkey.get_value('EnableLUA', as_json=self.as_json),
            'enable_virtualization': subkey.get_value('EnableVirtualization', as_json=self.as_json),
            'filter_admin_token': subkey.get_value('FilterAdministratorToken', as_json=self.as_json),
            'consent_prompt_admin': subkey.get_value('ConsentPromptBehaviorAdmin', as_json=self.as_json),
            'consent_prompt_user': subkey.get_value('ConsentPromptBehaviorUser', as_json=self.as_json)
        }


